package si.ijs.urbanite.simserver.network;

import org.jfree.data.json.impl.JSONObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.server.ResponseStatusException;
import si.ijs.urbanite.simserver.city.CityRepository;
import si.ijs.urbanite.simserver.entities.City;
import si.ijs.urbanite.simserver.entities.Network;
import si.ijs.urbanite.simserver.entities.Plan;
import si.ijs.urbanite.simserver.plan.PlanRepository;
import si.ijs.urbanite.simserver.utils.GeoJsonExport;
import org.matsim.run.NetworkCleaner;

import javax.validation.Valid;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.List;
import java.util.Optional;
import java.util.Scanner;

@RestController
@CrossOrigin
@RequestMapping("/traffic-simulation/network")
public class NetworkController {

    private final NetworkRepository networkRepository;
    private final NetworkCreationService networkCreationService;
    private final CityRepository cityRepository;
    private final PlanRepository planRepository;
    Logger logger = LoggerFactory.getLogger(NetworkController.class);

    public NetworkController(
            NetworkRepository networkRepository,
            NetworkCreationService networkCreationService,
            PlanRepository planRepository,
            CityRepository cityRepository) {
        this.networkRepository = networkRepository;
        this.networkCreationService = networkCreationService;
        this.cityRepository = cityRepository;
        this.planRepository = planRepository;
    }

    @GetMapping("")
    public Iterable<Network> getNetworks() {
        return networkRepository.findAll();
    }

    @GetMapping("/{netId}")
    public Network getNetwork(@PathVariable int netId) {
        Optional<Network> network = networkRepository.findById(netId);
        if (!network.isPresent())
            throw new ResponseStatusException(
                    HttpStatus.NOT_FOUND,
                    String.format("Network with the id %d does not exist.", netId));
        return network.get();
    }

    @GetMapping("/{netId}/plans")
    public Iterable<Plan> getPlansByNetwork(@PathVariable int netId) {
        Optional<Network> network = networkRepository.findById(netId);
        if (!network.isPresent())
            throw new ResponseStatusException(HttpStatus.NOT_FOUND, "Network with given id does not exist.");
        Iterable<Plan> plans = planRepository.findByNetwork(network.get());
        return plans;
    }

    @PostMapping("")
    public Network createNetwork(@RequestBody @Valid NetworkDTO networkDTO) {

        City city = cityRepository.findByCityId(networkDTO.cityId);
        if (city == null)
            throw new ResponseStatusException(HttpStatus.NOT_FOUND, "City with given id does not exists");
        if (networkDTO.name == null)
            networkDTO.name = networkDTO.cityId + " Network";
        if (networkDTO.description == null)
            networkDTO.description = "";

        Network network = networkRepository.save(new Network(networkDTO.name, networkDTO.description, city,
                networkDTO.north, networkDTO.south, networkDTO.east, networkDTO.west));
        networkCreationService.cutWithBoundingBox(network);
        return network;
    }

    @DeleteMapping("/{netId}")
    public ResponseEntity<Integer> deleteNetwork(@PathVariable int netId) {
        if (networkRepository.existsById(netId))
            networkRepository.deleteById(netId);
        else
            throw new ResponseStatusException(HttpStatus.NOT_FOUND, "Network with given id does not exists");
        return new ResponseEntity<>(netId, HttpStatus.OK);
    }

    @GetMapping("/validate/{city}/{networkId}")
    public String validateNetwork(@PathVariable String city, @PathVariable String networkId,
            @RequestHeader String[] links_array) throws IOException {

        Path path = Paths.get("data", city, "original_input", "networks", networkId, "network.xml");
        final NetworkCleaner cleaner = new NetworkCleaner();
        cleaner.run(String.valueOf(path), String.valueOf(path));
        List<String> links = Arrays.asList(links_array);

        File file = new File(String.valueOf(path));
        for (String link : links) {
            try {
                Scanner scanner = new Scanner(file);
                while (scanner.hasNextLine()) {
                    String line = scanner.nextLine();
                    if (line.contains("\"" + link + "\"")) {
                        links.remove(link);
                    }

                }
                scanner.close();
            } catch (FileNotFoundException e) {
                logger.error("file not found");
                // handle this
            }

        }
        return String.valueOf(links);
    }

    // TODO useless right now - Convert OSM file to Matsim
    @PutMapping("/{netId}/convertToMatsim")
    public Network convertToMatsim(@PathVariable int netId) {
        Optional<Network> network = networkRepository.findById(netId);
        if (!network.isPresent())
            throw new ResponseStatusException(HttpStatus.NOT_FOUND, "Network with given id does not exists");
        networkCreationService.convertToMatsim(network.get());
        return network.get();
    }

    @GetMapping("/geojson/{id}")
    public JSONObject getGeoJson(@PathVariable int id) {
        // get network
        Optional<Network> network = networkRepository.findById(id);
        File networkFile = new File(network.get().getNetworkFileOSM());
        return GeoJsonExport.networkToGeoJson(networkFile);
    }

}
