package si.ijs.urbanite.simserver.network;

import org.matsim.core.config.Config;
import org.matsim.core.config.ConfigUtils;
import org.matsim.pt2matsim.config.OsmConverterConfigGroup;
import org.matsim.pt2matsim.osm.lib.Osm;
import org.matsim.pt2matsim.run.Osm2MultimodalNetwork;
import org.openstreetmap.osmosis.core.Osmosis;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.messaging.simp.SimpMessagingTemplate;
import org.springframework.scheduling.annotation.Async;
import org.springframework.stereotype.Service;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;
import si.ijs.urbanite.simserver.entities.Network;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.*;
import java.util.logging.SimpleFormatter;
import java.util.logging.StreamHandler;

@Service
public class NetworkCreationService {

    private static Logger logger = LoggerFactory.getLogger(NetworkCreationService.class);

    private final NetworkRepository networkRepository;
    private final SimpMessagingTemplate simpMessagingTemplate;

    @Value("${urbanite.datadirectory}")
    private String dataPath;

    public NetworkCreationService(NetworkRepository networkRepository, SimpMessagingTemplate simpMessagingTemplate) {
        this.networkRepository = networkRepository;
        this.simpMessagingTemplate = simpMessagingTemplate;
    }

    @Async
    public void cutWithBoundingBox(Network network) {
        logger.info("Starting to process the network with Osmosis");

        try {
            Config config = ConfigUtils.createConfig();

            Path basePath = Paths.get(dataPath, network.getCity().getCityId());
            Path cityFile = basePath.resolve(Paths.get("original_input", "euskadi-latest.osm.pbf"));
            logger.info("\n\n\nbase_network.som.pbf file location:  " + cityFile.toAbsolutePath());
            Path networkFileOSM = basePath.resolve(Paths.get("original_input/networks", network.getNetworkFileOSM()));
            logger.info("networkFileOSM file location:  " + networkFileOSM.toAbsolutePath() + "\n\n\n");
            String command = String.format("osmosis --rb " + cityFile.toAbsolutePath() + " --bb top="
                    + network.getNorth() + " bottom=" + network.getSouth()
                    + " left=" + network.getWest() + " right=" + network.getEast()
                    + " --wx " + networkFileOSM.toAbsolutePath());

            // TODO fix this
            // could be done from cmd
            logger.info("\n\ncityfile\n\n" + command + "\n\n");
            Process process = Runtime.getRuntime().exec(command);
            logger.info("Done processing the network with Osmosis");
        } catch (Exception e) {
            logger.warn(String.format("There was an error processing the network: %s", e.getMessage()));
            network.setConversionError(e.getMessage());
        }

        network.setDoneOsmProcessing(true);
        networkRepository.save(network);
        if (network.getConversionError() == null) {
            this.sendProgress(network, "Finished extracting bounding box");
            this.convertToMatsim(network);
        } else {
            this.sendProgress(network, "Error while processing the network");
        }
    }

    @Async
    public void convertToMatsim(Network network) {
        logger.info("Starting to process the network with MATSim");
        Path basePath = Paths.get(dataPath, network.getCity().getCityId());
        Path networkFileOSM = basePath.resolve(Paths.get("networks", network.getNetworkFileOSM()));
        Path networkFile = basePath.resolve(Paths.get("networks", network.getNetworkFile()));

        OsmConverterConfigGroup config = createConfig();
        config.setOsmFile(networkFileOSM.toAbsolutePath().toString());
        config.setOutputNetworkFile(networkFile.toAbsolutePath().toString());

        Osm2MultimodalNetwork.run(config);
        network.setDoneMATSimProcessing(true);
        networkRepository.save(network);
        logger.info("Done processing the network with MATSim");
        this.sendProgress(network, "Finished converting to Matsim network");
    }

    public static OsmConverterConfigGroup createConfig() {
        OsmConverterConfigGroup defaultConfig = OsmConverterConfigGroup.createDefaultConfig();

        Set<String> bicycleSingleton = Collections.singleton("bicycle");
        Set<String> carSingleton = Collections.singleton("car");

        defaultConfig.addParameterSet(new OsmConverterConfigGroup.OsmWayParams(Osm.Key.HIGHWAY,
                CustomOsm.Value.CYCLEWAY, 1, 30.0 / 3.6, 1.0, 300, false, bicycleSingleton));
        defaultConfig.addParameterSet(new OsmConverterConfigGroup.OsmWayParams(Osm.Key.HIGHWAY, CustomOsm.Value.TRACK,
                1, 30.0 / 3.6, 1.0, 300, false, bicycleSingleton));
        defaultConfig.addParameterSet(new OsmConverterConfigGroup.OsmWayParams(Osm.Key.HIGHWAY, CustomOsm.Value.STEPS,
                1, 1.0 / 3.6, 1.0, 300, false, bicycleSingleton));
        defaultConfig.addParameterSet(new OsmConverterConfigGroup.OsmWayParams(Osm.Key.HIGHWAY,
                CustomOsm.Value.PEDESTRIAN, 1, 30.0 / 3.6, 1.0, 300, false, bicycleSingleton));
        defaultConfig.addParameterSet(new OsmConverterConfigGroup.OsmWayParams(Osm.Key.HIGHWAY, CustomOsm.Value.PATH, 1,
                20.0 / 3.6, 1.0, 300, false, bicycleSingleton));
        defaultConfig.addParameterSet(new OsmConverterConfigGroup.OsmWayParams(Osm.Key.HIGHWAY,
                CustomOsm.Value.FOOT_WAY, 1, 30.0 / 3.6, 1.0, 300, false, bicycleSingleton));
        defaultConfig.addParameterSet(new OsmConverterConfigGroup.OsmWayParams(Osm.Key.HIGHWAY, Osm.Value.SERVICE, 1,
                30.0 / 3.6, 1.0, 300, false, bicycleSingleton));

        defaultConfig.addParameterSet(new OsmConverterConfigGroup.RoutableSubnetworkParams("car", carSingleton));
        defaultConfig
                .addParameterSet(new OsmConverterConfigGroup.RoutableSubnetworkParams("bicycle", bicycleSingleton));
        defaultConfig.addParameterSet(new OsmConverterConfigGroup.RoutableSubnetworkParams("bicycle",
                new HashSet<>(Arrays.asList("car", "bicycle"))));
        defaultConfig.addParameterSet(new OsmConverterConfigGroup.RoutableSubnetworkParams("bus",
                new HashSet<>(Arrays.asList("car", "bus"))));

        return defaultConfig;
    }

    public static void checkIfOsmEmpty(File file)
            throws SAXException, ParserConfigurationException, IOException, OSMOutputFileEmpty {
        try {
            SAXParserFactory factory = SAXParserFactory.newInstance();
            SAXParser saxParser = factory.newSAXParser();
            saxParser.parse(file, new DefaultHandler() {
                private boolean moreThanOneTag = false;

                @Override
                public void startElement(String uri, String localName, String qName, Attributes attributes)
                        throws SAXException {
                    if (moreThanOneTag) {
                        throw new SAXTerminatorException();
                    } else {
                        moreThanOneTag = true;
                    }
                }
            });
        } catch (SAXTerminatorException e) {
            return;
        }
        throw new OSMOutputFileEmpty("Output OSM file is empty, boundary coordinates were most likely out of bounds");
    }

    public static class CustomOsm {
        public static final class Value {
            public final static String FOOT_WAY = "footway";
            public final static String STEPS = "steps";
            public final static String PATH = "path";
            public final static String PEDESTRIAN = "pedestrian";
            public final static String CYCLEWAY = "cycleway";
            public final static String TRACK = "track";
            public final static String LANE = "lane";
            public final static String OPPOSITE_LANE = "opposite_lane";
            public final static String OPPOSITE = "opposite";
            public final static String SHARED_LANE = "shared_lane";
        }
    }

    public static class SAXTerminatorException extends SAXException {
    }

    public static class OSMOutputFileEmpty extends Exception {
        public OSMOutputFileEmpty(String message) {
            super(message);
        }
    }

    public void sendProgress(Network n, String message) {
        this.simpMessagingTemplate.convertAndSend("/topic/networkProcessingStatus", n,
                Collections.singletonMap("conversionWSMessage", message));
    }

}
