package si.ijs.urbanite.simserver.network;

import org.matsim.core.config.Config;
import org.matsim.core.config.ConfigUtils;
import org.matsim.core.config.ConfigWriter;
import org.matsim.core.utils.collections.CollectionUtils;
import org.matsim.pt2matsim.config.OsmConverterConfigGroup;
import org.matsim.pt2matsim.config.PublicTransitMappingConfigGroup;
import org.matsim.pt2matsim.osm.OsmMultimodalNetworkConverter;
import org.matsim.pt2matsim.osm.lib.*;
import org.matsim.pt2matsim.run.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.matsim.pt2matsim.tools.NetworkTools;

import java.io.File;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

public class OsmConverter {
    Logger logger = LoggerFactory.getLogger(OsmConverter.class);

    public String city;
    public String epsg;
    public int networkId;
    // this class contains methods for converting osm network to matsim network
    private static String input;
    private static String output_original_input_folder;
    private static String output_network_folder;
    // private final String inter = example + "input_network/intermediate/";
    // private final String output = example + "input_network/output/";

    public void createNetwork(String epsg,
                              String city,
                              int networkId,
                              boolean gtfs,
                              boolean fine_grained,
                              String simulationName) {
        this.epsg = epsg;
        this.city = city;
        this.networkId = networkId;
        this.input = Paths.get("data",city, "before_simulation").toString();
        this.output_network_folder = Paths.get("data", city, "original_input", "networks", String.valueOf(networkId)).toString();
        this.output_original_input_folder = Paths.get("data", city, "original_input").toString();
        prepare();

        // 1. Convert a gtfs schedule to an unmapped transit schedule
        if (gtfs) {
            if (simulationName.equals("Messina PT simulation")){
                gtfsToSchedule("gtfs_lineAdded");
            }
            else {
                gtfsToSchedule("gtfs");
            }

        }
        // 2. Convert an osm map to a MATSim network
        // create a config file (or adjust an existing one by hand)
        createOsmConfigFile(Paths.get(input, "OsmConverterConfig.xml").toString(), gtfs, fine_grained);
        // Convert the OSM file using the config
        OsmConverterConfigGroup config = OsmConverterConfigGroup.loadConfig(Paths.get(input , "OsmConverterConfig.xml").toString());
        AllowedTagsFilter filter = new AllowedTagsFilter();
        filter.add(Osm.ElementType.WAY, Osm.Key.HIGHWAY, null);
        filter.add(Osm.ElementType.WAY, Osm.Key.HIGHWAY, "cycleway");
        filter.add(Osm.ElementType.WAY, Osm.Key.HIGHWAY, "footway");
        filter.add(Osm.ElementType.WAY, Osm.Key.HIGHWAY, "pedestrian");
        filter.add(Osm.ElementType.WAY, Osm.Key.HIGHWAY, "track");
        filter.add(Osm.ElementType.WAY, Osm.Key.HIGHWAY, Osm.Value.SERVICE);
        filter.add(Osm.ElementType.WAY, Osm.Key.HIGHWAY, "steps");

        /* done */
        OsmData osmData = new OsmDataImpl(filter);
        new OsmFileReader(osmData).readFile(config.getOsmFile());

        OsmMultimodalNetworkConverter converter = new OsmMultimodalNetworkConverter(osmData);
        converter.convert(config);

        NetworkTools.writeNetwork(converter.getNetwork(), config.getOutputNetworkFile());

        if (gtfs) {
            // 3. Map the schedule onto the network
            // create a config file (or adjust an existing one by hand)
            createMapperConfigFile(Paths.get(input, "MapperConfigAdjusted.xml").toString());
            // Map the schedule using the config
            PublicTransitMapper.main(new String[] { Paths.get(input, "MapperConfigAdjusted.xml").toString() });

            // 4. Do a plausibility check
            checkPlausibility();
        }
    }

    /** Create output folder if not existing **/
    public void prepare() {

        new File(Paths.get(output_network_folder).toString()).mkdirs();

    }

    /**
     * 1. A GTFS or HAFAS Schedule or a OSM map with information on public transport
     * has to be converted to an unmapped MATSim Transit Schedule.
     *
     * Here as a first example, the GTFS-schedule of GrandRiverTransit,
     * Waterloo-Area, Canada, is converted.
     */
    public void gtfsToSchedule(String gtfs) {


        String[] gtfsConverterArgs = new String[] {
                // [0] gtfs zip file
                Paths.get(input, gtfs).toString(),
                // [1] which service ids should be used. One of the following:
                // dayWithMostTrips, date in the format yyyymmdd, , dayWithMostServices, all
                "dayWithMostTrips",
                // [2] the output coordinate system. Use WGS84 for no transformation.
                epsg,
                // [3] output transit schedule file
                Paths.get(input, "schedule_unmapped.xml").toString(),
                // [4] output default vehicles file (optional)
                Paths.get(output_original_input_folder,"transit_vehicles.xml").toString()
        };
        Gtfs2TransitSchedule.main(gtfsConverterArgs);
    }

    /**
     * Here as a second example, the HAFAS-schedule of the BrienzRothornBahn,
     * Switzerland, is converted.
     */

    /**
     * 2. A MATSim network of the area is required. If no such network is already
     * available, the PT2MATSim package provides the possibility to use OSM-maps as
     * data-input.
     *
     * Here as an example, the OSM-extract of the city centre of Waterloo, Canada,
     * is converted.
     */
    public void createOsmConfigFile(String configFile, boolean gtfs, boolean fine_grained) {
        // Create a default createOsmConfigFile-Config:
        CreateDefaultOsmConfig.main(new String[] { Paths.get(input, "OsmConverterConfigDefault.xml").toString() });

        // Open the createOsmConfigFile Config and set the parameters to the required
        // values
        // (usually done manually by opening the config with a simple editor)
        Config osmConverterConfig = ConfigUtils.loadConfig(Paths.get(input, "OsmConverterConfigDefault.xml").toString(),
                new OsmConverterConfigGroup());

        OsmConverterConfigGroup osmConfig = ConfigUtils.addOrGetModule(osmConverterConfig,
                OsmConverterConfigGroup.class);

        Set<String> bicycleSingleton = Collections.singleton("bicycle");
        Set<String> ferrySingleton = Collections.singleton("ferry");
        Set<String> carSingleton = Collections.singleton("car");
        osmConfig.addParameterSet(new OsmConverterConfigGroup.OsmWayParams(Osm.Key.HIGHWAY, "cycleway", 1, 30.0 / 3.6, 1.0, 300, false, bicycleSingleton));
        if (fine_grained) {
//            osmConfig.addParameterSet(new OsmConverterConfigGroup.OsmWayParams(Osm.Key.HIGHWAY, "cycleway", 1, 30.0 / 3.6, 1.0, 300, false, bicycleSingleton));
            osmConfig.addParameterSet(new OsmConverterConfigGroup.OsmWayParams(Osm.Key.HIGHWAY, "track", 1, 30.0 / 3.6, 1.0, 300, false, bicycleSingleton));
//            osmConfig.addParameterSet(new OsmConverterConfigGroup.OsmWayParams(Osm.Key.HIGHWAY, "steps", 1, 1.0 / 3.6, 1.0, 300, false, bicycleSingleton));
            osmConfig.addParameterSet(new OsmConverterConfigGroup.OsmWayParams(Osm.Key.HIGHWAY, "pedestrian", 1, 30.0 / 3.6, 1.0, 300, false, bicycleSingleton));
            osmConfig.addParameterSet(new OsmConverterConfigGroup.OsmWayParams(Osm.Key.HIGHWAY, "path", 1, 20.0 / 3.6, 1.0, 300, false, bicycleSingleton));
            osmConfig.addParameterSet(new OsmConverterConfigGroup.OsmWayParams(Osm.Key.HIGHWAY, "footway", 1, 30.0 / 3.6, 1.0, 300, false, bicycleSingleton));
            osmConfig.addParameterSet(new OsmConverterConfigGroup.OsmWayParams(Osm.Key.HIGHWAY, "service", 1, 30.0 / 3.6, 1.0, 300, false, bicycleSingleton));
        }
        //for the ferries (Helsinki)
        osmConfig.addParameterSet(new OsmConverterConfigGroup.OsmWayParams(Osm.Key.ROUTE,Osm.Value.FERRY,1, 30.0/3.6, 1.0,300,false,ferrySingleton));
        osmConfig.addParameterSet(new OsmConverterConfigGroup.OsmWayParams("amenity","ferry_terminal",1, 30.0/3.6, 1.0,300,false,ferrySingleton));

        //		=====================================================

        osmConfig.addParameterSet(new OsmConverterConfigGroup.RoutableSubnetworkParams("car", carSingleton));
        osmConfig.addParameterSet(new OsmConverterConfigGroup.RoutableSubnetworkParams("bicycle", bicycleSingleton));
        osmConfig.addParameterSet(new OsmConverterConfigGroup.RoutableSubnetworkParams("bicycle", new HashSet<>(Arrays.asList("car", "bicycle"))));
        osmConfig.addParameterSet(new OsmConverterConfigGroup.RoutableSubnetworkParams("bus", new HashSet<>(Arrays.asList("car", "bus"))));


        osmConfig.setOsmFile(Paths.get(input, "network.osm").toString());
        osmConfig.setOutputCoordinateSystem(epsg);
        if (!gtfs) {
            osmConfig.setOutputNetworkFile(Paths.get(output_network_folder, "network.xml").toString());
        } else {
            osmConfig.setOutputNetworkFile(Paths.get(input, "network.xml").toString());
        }

        // Save the createOsmConfigFile config (usually done manually)
        new ConfigWriter(osmConverterConfig).write(configFile);
    }

    /**
     * 3. The core of the PT2MATSim-package is the mapping process of the schedule
     * to the network.
     *
     * Here as an example, the unmapped schedule of GrandRiverTransit (previously
     * converted from GTFS) is mapped to the converted OSM network of the Waterloo
     * Area, Canada.
     */
    public void createMapperConfigFile(String configFile) {
        // Create a mapping config:
        CreateDefaultPTMapperConfig.main(new String[] { Paths.get(input, "MapperConfigDefault.xml").toString() });
        // Open the mapping config and set the parameters to the required values
        // (usually done manually by opening the config with a simple editor)
        Config config = ConfigUtils.loadConfig(Paths.get(input, "MapperConfigDefault.xml").toString(),
                PublicTransitMappingConfigGroup.createDefaultConfig());
        PublicTransitMappingConfigGroup ptmConfig = ConfigUtils.addOrGetModule(config,
                PublicTransitMappingConfigGroup.class);
        ptmConfig.getModesToKeepOnCleanUp().add("car,bicycle");

        ptmConfig.setInputNetworkFile(Paths.get(input, "network.xml").toString());
        ptmConfig.setOutputNetworkFile(Paths.get(output_network_folder, "network.xml").toString());
        ptmConfig.setOutputScheduleFile(Paths.get(output_original_input_folder, "schedule.xml").toString());

//        File street_network_folder_path = new File(Paths.get(output_network_folder, "osm_converter").toString());
//        street_network_folder_path.mkdirs();
//
//        ptmConfig.setOutputStreetNetworkFile(Paths.get(street_network_folder_path.toString(), "streetnetwork.xml").toString());
        ptmConfig.setInputScheduleFile(Paths.get(input, "schedule_unmapped.xml").toString());
        ptmConfig.setScheduleFreespeedModes(CollectionUtils.stringToSet("rail, light_rail,ferry,tram"));
        // Save the mapping config
        // (usually done manually)
        new ConfigWriter(config).write(configFile);
    }

    /**
     * 4. The PT2MATSim package provides a plausibility checker to get quick
     * feedback on the mapping process.
     *
     * Here as an example, the mapped transit schedule and the multimodal network
     * created in step 3 is checked for plausibility.
     */
    public void checkPlausibility() {
        CheckMappedSchedulePlausibility.run(Paths.get(output_original_input_folder, "schedule.xml").toString(),
                Paths.get(output_network_folder, "network.xml").toString(), epsg, Paths.get(output_network_folder, "plausibilityResults").toString());
    }

    public static void main(String[] args) {
        OsmConverter osm = new OsmConverter();
        osm.createNetwork("EPSG:32630", "bilbao", 2, true, true,"xx");
    }
}
