package si.ijs.urbanite.simserver.scripts;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.messaging.simp.SimpMessagingTemplate;
import org.springframework.stereotype.Service;
import si.ijs.urbanite.simserver.entities.Network;
import si.ijs.urbanite.simserver.entities.Plan;
import si.ijs.urbanite.simserver.entities.Simulation;
import si.ijs.urbanite.simserver.network.NetworkRepository;
import si.ijs.urbanite.simserver.plan.PlanRepository;
import si.ijs.urbanite.simserver.simulation.SimulationRepository;
import si.ijs.urbanite.simserver.simulation.SimulationRunService;
import si.ijs.urbanite.simserver.simulation.SimulationWithParamsDTO;
import si.ijs.urbanite.simserver.travel_demand.TravelDemandAmsterdam;
import si.ijs.urbanite.simserver.travel_demand.TravelDemandGenerationBilbao;
import si.ijs.urbanite.simserver.travel_demand.TravelDemandGenerationHelsinki;

import java.io.*;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Optional;

@Service
public class ScriptsService {
    private final static Logger logger = LoggerFactory.getLogger(SimulationRunService.class);
    private final SimpMessagingTemplate simpMessagingTemplate;
    private final PlanRepository planRepository;
    private final NetworkRepository networkRepository;
    private final SimulationRepository simulationRepository;

    // TODO: Implement messenging to user when script execution is finished
    public ScriptsService(SimpMessagingTemplate simpMessagingTemplate,
                          PlanRepository planRepository, NetworkRepository networkRepository, SimulationRepository simulationRepository) {
        this.simpMessagingTemplate = simpMessagingTemplate;
        this.planRepository = planRepository;
        this.networkRepository = networkRepository;
        this.simulationRepository = simulationRepository;
    }

    public int runPreprocessing(String city_name) throws IOException {
        String line;
        int error_flag = 0;
        Path pathToPythonPreprocess = Paths.get("population_generation", city_name + "_preprocess.py");
        BufferedReader bfr;

        // run python preprocessing
        ProcessBuilder pb = new ProcessBuilder("python3", pathToPythonPreprocess.toString());
        Process process = pb.start();
        bfr = new BufferedReader(new InputStreamReader(process.getErrorStream()));
        logger.info("Output error stream for python preprocess script:\n");
        while ((line = bfr.readLine()) != null) {
            logger.error(line);
            error_flag = 1;
        }
        if (error_flag == 1) return -1;
        logger.info("FINISHED python");
        return 0;
    }

    public int runGeneratePopulation(String city_name) throws IOException {
        String line;
        int error_flag = 0;

        Path pathToRScript = Paths.get("population_generation", city_name + "_population.R");
        BufferedReader bfr;

        Process process;
        // run R script (synthetic population generation)
        ProcessBuilder pb = new ProcessBuilder("Rscript", pathToRScript.toString());
        process = pb.start();
        bfr = new BufferedReader(new InputStreamReader(process.getErrorStream()));
        logger.info("Output error stream for R script:\n");
        while ((line = bfr.readLine()) != null) {
            logger.error(line);
            error_flag = 1;

        }
        if (error_flag == 1) return -1;

        logger.info("FINISHED R");
        return 0;

    }

    public int runTravelDemand(String city_name, String net_id, String sim_id, String plan_name, SimulationWithParamsDTO paramsDTO) throws IOException {
        Optional<Network> net_opt = networkRepository.findById(Integer.valueOf(net_id));
        Optional<Simulation> sim_opt = simulationRepository.findById(Integer.valueOf(sim_id));
        Network net;
        Simulation sim;
        if (net_opt.isPresent() && sim_opt.isPresent()) {
            net = net_opt.get();
            sim = sim_opt.get();
        }
        else {
            logger.error("Network or Simulation do not exist. Stopping travel demand generation.");
            return -1;
        }
        Plan plan = planRepository.save(new Plan(plan_name, net));
        String plan_id = String.valueOf(plan.getId());
        try {
            switch (city_name) {
                case "amsterdam": {
                    //use TravelDemandAmsterdamCyclistsCut instead
                    TravelDemandAmsterdam.generateTravelDemand(net_id, plan_id);
                    break;
                }
                case "bilbao": {
                    TravelDemandGenerationBilbao.generateTravelDemand(net_id, plan_id, paramsDTO);
                    break;
                }
                case "helsinki": {
                    //use TravelDemandHelsinkiCut instead
                    TravelDemandGenerationHelsinki.generateTravelDemand(net_id, plan_id, paramsDTO);
                    break;
                }
                case "messina": {
                    break;
                }
            }
        } catch (Exception e) {
            StringWriter sw = new StringWriter();
            PrintWriter pw = new PrintWriter(sw);
            e.printStackTrace(pw);
            logger.error(sw.toString());
            return -1;
        }
        logger.info("Travel demand generation finished");
        return 0;
    }
}
