package si.ijs.urbanite.simserver.simulation;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.http.HttpStatus;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.server.ResponseStatusException;
import si.ijs.urbanite.simserver.city.CityRepository;
import si.ijs.urbanite.simserver.entities.Network;
import si.ijs.urbanite.simserver.entities.Plan;
import si.ijs.urbanite.simserver.entities.Simulation;
import si.ijs.urbanite.simserver.entities.City;
import si.ijs.urbanite.simserver.network.NetworkRepository;
import si.ijs.urbanite.simserver.plan.PlanRepository;

import javax.validation.Valid;
import java.io.FileNotFoundException;
import java.util.HashMap;
import java.util.Optional;

@RestController()
@CrossOrigin
@RequestMapping("/traffic-simulation/simulation")
public class SimulationController {

    private final static Logger logger = LoggerFactory.getLogger(SimulationController.class);
    private final static HashMap<String, String[]> cityToData = new HashMap<>();

    static {
        cityToData.put("AMST", new String[]{"population_10000.xml", "vehicles_unmapped.xml"});
        cityToData.put("BILBAO", new String[]{"bilbao-current-population.xml", "bilbo-current-vehicles.xml"});
    }

    private final SimulationRepository simulationRepository;
    private final NetworkRepository networkRepository;
    private final PlanRepository planRepository;
    private final SimulationRunService simulationRunService;
    private final CityRepository cityRepository;

    @Value("${urbanite.datadirectory}")
    private String dataPath;
    @Value("${urbanite.city}")
    private String cityId;

    public SimulationController(SimulationRepository simulationRepository,
                                NetworkRepository networkRepository,
                                SimulationRunService simulationRunService,
                                PlanRepository planRepository, CityRepository cityRepository) {
        this.simulationRepository = simulationRepository;
        this.networkRepository = networkRepository;
        this.simulationRunService = simulationRunService;
        this.planRepository = planRepository;
        this.cityRepository = cityRepository;

        this.dataPath = this.dataPath + this.cityId;
    }

    @GetMapping("")
    public Iterable<Simulation> getSimulations() {
        return simulationRepository.findAll();
    }

    @GetMapping("/{id}")
    public Simulation getSimulation(@PathVariable int id) throws FileNotFoundException {
        Optional<Simulation> simulation = simulationRepository.findById(id);
        if (!simulation.isPresent())
            throw new ResponseStatusException(HttpStatus.NOT_FOUND, "Simulation with given id does not exists");
        return simulation.get();
    }

    @GetMapping("/{id}/network")
    public Network getSimulationNetwork(@PathVariable int id) throws FileNotFoundException {
        Optional<Simulation> simulation = simulationRepository.findById(id);
        if (!simulation.isPresent())
            throw new ResponseStatusException(HttpStatus.NOT_FOUND, "Simulation with given id does not exists");
        return simulation.get().getPlan().getNetwork();
    }

    @PostMapping("")
    public Simulation postSimulation(@RequestBody @Valid SimulationDTO simulationDTO) {

        Optional<City> cityOpt = cityRepository.findById(simulationDTO.city);
        Optional<Plan> planOpt = planRepository.findById(simulationDTO.planId);

        if (cityOpt.isPresent() && planOpt.isPresent()){
            Simulation simulation = new Simulation(simulationDTO.name, simulationDTO.description, planOpt.get(), cityOpt.get());
            simulationRepository.save(simulation);
            return simulation;
        }
        return null;

    }

    @PostMapping("/withParams/run")
    public Simulation runSimulationWithParams(@RequestBody @Valid SimulationWithParamsDTO simWithParams) {
        Optional<Simulation> simulationOptional = this.simulationRepository.findById(simWithParams.simId);
        if (!simulationOptional.isPresent())
            throw new ResponseStatusException(HttpStatus.NOT_FOUND, "Simulation with given id does not exists");
        Simulation simulation = simulationOptional.get();
        simulationRunService.runSimulation(simulation, simWithParams);
        return simulation;
    }
    @GetMapping("/{id}/run")
    public Simulation runSimulation(@PathVariable int id) {
        logger.info("Run simulation " + id + " called");
        Optional<Simulation> simulationOptional = this.simulationRepository.findById(id);
        if (!simulationOptional.isPresent())
            throw new ResponseStatusException(HttpStatus.NOT_FOUND, "Simulation with given id does not exists");
        logger.info("Simulation with given id found");
        Simulation simulation = simulationOptional.get();
        simulationRunService.runSimulation(simulation, null);

        return simulation;
    }

}
