package si.ijs.urbanite.simserver.simulation;

import ch.sbb.matsim.routing.pt.raptor.SwissRailRaptorModule;
import org.joda.time.DateTime;
import org.matsim.api.core.v01.Scenario;
import org.matsim.api.core.v01.network.Network;
import org.matsim.api.core.v01.network.NetworkWriter;
import org.matsim.contrib.cadyts.general.CadytsConfigGroup;
import org.matsim.contrib.emissions.utils.EmissionsConfigGroup;
import org.matsim.core.config.Config;
import org.matsim.core.config.ConfigUtils;
import org.matsim.core.config.groups.ControlerConfigGroup;
import org.matsim.core.controler.Controler;
import org.matsim.core.controler.OutputDirectoryHierarchy;
import org.matsim.core.network.NetworkUtils;
import org.matsim.core.scenario.ScenarioUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.messaging.simp.SimpMessagingTemplate;
import org.springframework.scheduling.annotation.Async;
import org.springframework.stereotype.Service;
import si.ijs.urbanite.simserver.district_closer.DistrictCloserParams;
import si.ijs.urbanite.simserver.emissions.EmissionsOffline;
import si.ijs.urbanite.simserver.entities.Plan;
import si.ijs.urbanite.simserver.entities.Simulation;
import si.ijs.urbanite.simserver.network.AddNetowrkLayer;
import si.ijs.urbanite.simserver.network.AddTunnel;
import si.ijs.urbanite.simserver.network.NetworkChangeEvent;
import si.ijs.urbanite.simserver.network.OsmConverter;
import si.ijs.urbanite.simserver.plan.PlanRepository;
import si.ijs.urbanite.simserver.travel_demand.*;
import si.ijs.urbanite.simserver.utils.HelperFunctions;
import si.ijs.urbanite.simserver.vehicles.VehiclesGeneration;

import java.io.*;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Collections;
import java.util.List;


import static si.ijs.urbanite.simserver.emissions.EmissionsOffline.calculateEmissions;

@Service
public class SimulationRunService {

    private final static Logger logger = LoggerFactory.getLogger(SimulationRunService.class);
    private final static String ferriesPath = "data/helsinki/before_simulation/ferriesJatkasaari.csv";
    private final static String ferriesToAddPath = "data/helsinki/before_simulation/ferries_to_add.csv";
    private final SimpMessagingTemplate simpMessagingTemplate;
    private final SimulationRepository simulationRepository;
    private final PlanRepository planRepository;

    @Value("${urbanite.datadirectory}")
    private String dataPath;
    @Value("${urbanite.city}")
    private String city;

    public SimulationRunService(
            SimpMessagingTemplate simpMessagingTemplate,
            SimulationRepository simulationRepository,
            PlanRepository planRepository) {
        this.simpMessagingTemplate = simpMessagingTemplate;
        this.simulationRepository = simulationRepository;
        this.planRepository = planRepository;
    }

    public void runStepsBeforeSimulation(int networkId, int planId, String simulationName, SimulationWithParamsDTO paramsDTO) throws IOException {
        logger.info("running steps before simulation...");
        runPythonPreprocess();
        runRScript();
        if (this.city.equals("helsinki") && new File(ferriesToAddPath).exists()){
            String new_ferries = new String(Files.readAllBytes(Paths.get(ferriesToAddPath)), StandardCharsets.UTF_8);
            FileWriter fw = new FileWriter(ferriesPath,true);
            fw.write(new_ferries);
            fw.close();
        }
        generateTravelDemand(networkId, planId, simulationName, paramsDTO);

        generateNetwork(city,networkId,simulationName, String.valueOf(planId));
        logger.info("steps before simulation done.");
    }

    public void runPythonPreprocess() throws IOException {
        logger.info("Running python preproces");
        Path pathToPythonPreprocess = Paths.get("population_generation", this.city + "_preprocess.py");
        ProcessBuilder pb = new ProcessBuilder("python3", pathToPythonPreprocess.toString());
        Process process = pb.start();
        BufferedReader bfr = new BufferedReader(new InputStreamReader(process.getErrorStream()));
        HelperFunctions.logExternalProcess(logger, bfr, "python script (preprocess)");
    }

    public void runRScript() throws IOException {
        logger.info("Running R script");
        Path pathToRScript = Paths.get("population_generation", this.city + "_population.R");
        ProcessBuilder pb = new ProcessBuilder("Rscript", pathToRScript.toString());
        Process process = pb.start();
        BufferedReader bfr = new BufferedReader(new InputStreamReader(process.getErrorStream()));
        HelperFunctions.logExternalProcess(logger, bfr, "R script (population generation)");
    }

    public void generateTravelDemand(int networkId, int planId, String simulationName, SimulationWithParamsDTO params) throws IOException {
        logger.info("Creating results dir for travel demand");
        Path path = Paths.get(dataPath, city, "original_input", "networks", String.valueOf(networkId), "plans",
                String.valueOf(planId));
        if (!Files.exists(path)) {
            boolean result = new File(String.valueOf(path)).mkdirs();
            if (!result) {
                logger.error("Could not create travel demand path.");
                return;
            } else {
                logger.info(String.format("Create travel demand results dir: %s", path));
            }
        }

        logger.info("Generating travel demand.");
        switch (this.city) {
            case "amsterdam": {
                if (simulationName.equals("Amsterdam Noord simulation")){
                    TravelDmenadAmsterdamCyclistsCut.generateTravelDemand(String.valueOf(networkId), String.valueOf(planId), 1,true, params);
                }
                else {
                    TravelDmenadAmsterdamCyclistsCut.generateTravelDemand(String.valueOf(networkId), String.valueOf(planId), 1, false, params);
                }
                break;
            }
            case "bilbao": {
                TravelDemandGenerationBilbao.generateTravelDemand(String.valueOf(networkId), String.valueOf(planId), params);
                break;
            }
            case "helsinki": {
                TravelDemandGenerationHelsinkiCut.generateTravelDemand(String.valueOf(networkId), String.valueOf(planId));
                break;
            }
            case "messina": {
                TravelDemandGenerationMessina.generateTravelDemand(String.valueOf(networkId), String.valueOf(planId), params);
                break;
            }
        }
        logger.info("Travel demand generation finished");
    }
    public void generateNetwork(String city, int networkId, String simulationName, String plan_id){
        logger.info("Creating network");
        OsmConverter osm = new OsmConverter();
        switch (this.city) {
            case "amsterdam": {
                // no gtfs used
                osm.createNetwork("EPSG:32631",city,networkId,false, true, simulationName);
                break;
            }
            case "bilbao":
            {
                osm.createNetwork("EPSG:32630",city, networkId,true, true, simulationName);
                if (simulationName.matches("(.*)Bilbao Moyua LTZ(.*)")) {
                    logger.info("Simulation: Bilbao Moyua LTZ running");
                    Config config = ConfigUtils.loadConfig(Paths.get(dataPath,city,"original_input","config.xml").toString());
                    config.network().setInputFile(Paths.get(dataPath,city,"original_input","networks",String.valueOf(networkId),"network.xml").toString());
                    config.plans().setInputFile(Paths.get(dataPath,city,"original_input","networks",String.valueOf(networkId),"plans", plan_id, "plans.xml").toString());
                    Scenario sc = ScenarioUtils.loadScenario(config);
                    Network network = AddNetowrkLayer.addLayer(sc.getNetwork());
                    NetworkWriter writer = new NetworkWriter(network);
                    writer.write(Paths.get(dataPath,city,"original_input","networks",String.valueOf(networkId),"network.xml").toString());
                }
                break;
            }
            case "helsinki":{
                osm.createNetwork("EPSG:32634",city, networkId,true,false,simulationName);

                //if scenario 1 is selected: add tunnel
                if (simulationName.equals("Helsinki harbor tunnel simulation")){
                    AddTunnel.addTunnel(NetworkUtils.readNetwork(Paths.get(dataPath,city,"original_input","networks",String.valueOf(networkId),"network.xml").toString()), city, String.valueOf(networkId));
                }

                break;
            }
            case "messina": {
                osm.createNetwork("EPSG:32633",city, networkId,true, true, simulationName);
                break;
            }
        }
        logger.info("Travel demand generation finished");

    }

    public Config chooseCorrespondingConfigFile(String city_name, String simulation_id) {
        Config config = null;
        String config_path = Paths.get("data", city_name, "simulations", simulation_id, "config.xml").toString();
        config = ConfigUtils.loadConfig(config_path, new CadytsConfigGroup(), new EmissionsConfigGroup());

        if (config == null) {
            throw new IllegalArgumentException(
                    "config variable can not be null. Most probably it did not get initialized");
        }
        return config;
    }

    void copyFilesToEmissionsFolder(Path emissions_path, String simulated_city, String sim_id) throws IOException {

        Path simulation_path = Paths.get("data", simulated_city, "simulations", sim_id);

        List<Path> network_files = HelperFunctions.findFilesThatMatches(simulation_path, new String[] { "network.xml" });
        List<Path> vehicles_files = HelperFunctions.findFilesThatMatches(simulation_path, new String[] { "vehicles.xml" });
        List<Path> transit_vehicles_files = HelperFunctions.findFilesThatContain(simulation_path,
                new String[] { "transit" });
        List<Path> plans_files = HelperFunctions.findFilesThatContain(simulation_path,
                new String[] { "plans", "traveldemand" });

        String plans_destination = Paths.get(emissions_path.toString(), plans_files.get(0).getFileName().toString())
                .toString();
        HelperFunctions.copyFile(plans_files.get(0).toString(), plans_destination);

        if (!vehicles_files.isEmpty()) {
            String vehicles_destination = Paths
                    .get(emissions_path.toString(), vehicles_files.get(0).getFileName().toString()).toString();
            HelperFunctions.copyFile(vehicles_files.get(0).toString(), vehicles_destination);
        }

        if(!transit_vehicles_files.isEmpty()){
            String transit_vehicles_destination = Paths
                    .get(emissions_path.toString(), transit_vehicles_files.get(0).getFileName().toString()).toString();
            HelperFunctions.copyFile(transit_vehicles_files.get(0).toString(), transit_vehicles_destination);
        }

        String network_destination = Paths.get(emissions_path.toString(), network_files.get(0).getFileName().toString())
                .toString();
        HelperFunctions.copyFile(network_files.get(0).toString(), network_destination);
    }

    public void copyFiles(String network_id, String plans_id, String simulated_city, String sim_id) throws IOException {
        String simulation_path = Paths.get("data", simulated_city, "simulations", sim_id).toString();
        Path original_input_path = Paths.get("data", simulated_city, "original_input");
        Path emission_path = Paths.get("data", simulated_city, "input_emissions");

        // variables are in plural but we only expect to (find) return a single file
        // (except the vehicles files and config, we expect two)
        List<Path> network_files = HelperFunctions.findFilesThatMatches(
                Paths.get(original_input_path.toString(), "networks", network_id), new String[] { "network.xml" });
        List<Path> config_files = HelperFunctions.findFilesThatContain(original_input_path, new String[] { "config" });
        List<Path> vehicles_files = HelperFunctions.findFilesThatMatches(original_input_path,
                new String[] { "vehicles.xml" });
        List<Path> transit_vehicles_files = HelperFunctions.findFilesThatContain(original_input_path,
                new String[] { "transit" });
        List<Path> plans_files = HelperFunctions.findFilesThatContain(
                Paths.get(original_input_path.toString(), "networks", network_id, "plans", plans_id),
                new String[] { "plans", "traveldemand" });
        List<Path> schedule_files = HelperFunctions.findFilesThatContain(original_input_path,
                new String[] { "schedule.xml" });
        List<Path> counts_files = HelperFunctions.findFilesThatContain(original_input_path, new String[] { "count" });

        List<Path> hot_emission_files = HelperFunctions.findFilesThatContain(emission_path, new String[] { "hot" });
        List<Path> cold_emission_files = HelperFunctions.findFilesThatContain(emission_path, new String[] { "cold" });

        // We are only interested in taking one file and copying it to the destination
        String plans_destination = Paths.get(simulation_path, plans_files.get(0).getFileName().toString()).toString();
        HelperFunctions.copyFile(plans_files.get(0).toString(), plans_destination);

        String network_destination = Paths.get(simulation_path, network_files.get(0).getFileName().toString())
                .toString();
        HelperFunctions.copyFile(network_files.get(0).toString(), network_destination);
        for (Path p : config_files) {
            // to copy both config and emissions_config
            String destination = Paths.get(simulation_path, p.getFileName().toString()).toString();
            HelperFunctions.copyFile(p.toString(), destination);
        }

        // these files are not present on every city
        if (!schedule_files.isEmpty()) {
            String schedule_destination = Paths.get(simulation_path, schedule_files.get(0).getFileName().toString())
                    .toString();
            HelperFunctions.copyFile(schedule_files.get(0).toString(), schedule_destination);
        }

        if (!vehicles_files.isEmpty()) {
            String vehicles_destination = Paths.get(simulation_path, vehicles_files.get(0).getFileName().toString())
                    .toString();
            logger.error("VEHICLES DESTINATION:" + vehicles_destination);
            HelperFunctions.copyFile(vehicles_files.get(0).toString(), vehicles_destination);
        }

        if(!transit_vehicles_files.isEmpty()){
            String transit_vehicles_destination = Paths
                    .get(simulation_path, transit_vehicles_files.get(0).getFileName().toString()).toString();
            logger.error("TRANSIT VEHICLES DESTINATION:" + transit_vehicles_destination);

            HelperFunctions.copyFile(transit_vehicles_files.get(0).toString(), transit_vehicles_destination);
        }

        if (!counts_files.isEmpty()) {
            String counts_destination = Paths.get(simulation_path, counts_files.get(0).getFileName().toString())
                    .toString();
            HelperFunctions.copyFile(counts_files.get(0).toString(), counts_destination);
        }
        if (!hot_emission_files.isEmpty()) {
            String hot_emission_files_destination = Paths
                    .get(simulation_path, hot_emission_files.get(0).getFileName().toString()).toString();
            HelperFunctions.copyFile(hot_emission_files.get(0).toString(), hot_emission_files_destination);
        }
        if (!cold_emission_files.isEmpty()) {
            String cold_emission_files_destination = Paths
                    .get(simulation_path, cold_emission_files.get(0).getFileName().toString()).toString();
            HelperFunctions.copyFile(cold_emission_files.get(0).toString(), cold_emission_files_destination);
        }
    }

    @Async
    public void runSimulation(Simulation simulation, SimulationWithParamsDTO params) {
        logger.info(String.format("Run simulation called for simulation %d", simulation.getId()));
        this.city = simulation.getCity().getCityId();
        logger.info(String.format("Simulation in %s", city));


        int simulationId = simulation.getId();
        int networkId = simulation.getPlan().getNetwork().getId();
        Plan plan = getPlanForSimulation(simulation);
        int planId = plan.getId();
        String simulationName= simulation.getName();

        try {
            runStepsBeforeSimulation(networkId, planId,simulationName, params);
        } catch (IOException e) {
            logger.error("Error in steps before simulation. Shutting down");
            e.printStackTrace();
            return;
        }

        Path resultsPath = createResultDir(simulationId);
        try {
            copyFiles(String.valueOf(networkId), String.valueOf(planId), city, String.valueOf(simulationId));
        } catch (IOException ex) {
            logger.error("Error copying simulation files. Shutting down");
            ex.printStackTrace();
            return;
        }
        logger.info("Copy simulation files finished");
        try {
            copyFilesToEmissionsFolder(Paths.get("data", city, "input_emissions"), city, String.valueOf(simulationId));
        } catch (IOException ex) {
            logger.error("Error copying files to emissions folder. Shutting down");
            ex.printStackTrace();
            return;
        }
        logger.info("Copy files to emissions folder finished");

        sendProgress(simulation, Simulation.Status.RUNNING, "Preparing simulation");
        Config config = chooseCorrespondingConfigFile(city, String.valueOf(simulationId));

        //if bilbao close Moyua network scenario mis running, uncomment
//        config.network().setTimeVariantNetwork(true);

        Scenario sc = ScenarioUtils.loadScenario(config);


        Controler controler = new Controler(sc);
        switch (this.city) {
            case "amsterdam": {
                 controler = VehiclesGeneration.addBicycleTypes(config);
                 controler.getConfig().transit().setUseTransit(false);
            }
            case "helsinki":{
                controler.getConfig().transit().setUseTransit(false);
            }
        }
        //if bilbao close Moyua network scenario mis running, uncomment
        if (simulationName.matches("(.*)LTZ(.*)")){
            NetworkChangeEvent.createNetworkChangeEventClose(0,sc.getNetwork(),sc);
        }

        //if start time and end time are not empty. run bilbao start/end time of closure simualation
        if (params.closure_duration_time != 0 && params.closure_duration_time !=0){
            NetworkChangeEvent.createNetworkChangeEventClose(params.closure_start_time,sc.getNetwork(),sc);
            NetworkChangeEvent.createNetworkChangeEventOpen(params.closure_duration_time+ params.closure_start_time,sc.getNetwork(),sc);
        }

        controler.getConfig().controler()
                .setOverwriteFileSetting(OutputDirectoryHierarchy.OverwriteFileSetting.deleteDirectoryIfExists);
        controler.getConfig().controler().setOutputDirectory(resultsPath.toAbsolutePath().toString());
        controler.getConfig().controler().setCompressionType(ControlerConfigGroup.CompressionType.none);
        controler.getConfig().controler()
                .setEventsFileFormats(Collections.singleton(ControlerConfigGroup.EventsFileFormat.xml));

        controler.getConfig().controler().setWriteEventsInterval(5);
        controler.getConfig().controler().setLastIteration(50);
        controler.getConfig().network().setInputFile(Paths.get(city,"original_input","networks",String.valueOf(networkId),"network.xml").toString());
        controler.getConfig().plans().setInputFile(Paths.get(city,"original_input","networks",String.valueOf(networkId),"plans",String.valueOf(planId),"plans.xml").toString());

        sendProgress(simulation, Simulation.Status.RUNNING, "Simulation running");

        //// include cadyts into the plan scoring (this will add the cadyts corrections
        //// to the scores):
        // Cadyts will be included when we have more data
        /*
         * controler.addOverridingModule(new CadytsCarModule());
         *
         * controler.setScoringFunctionFactory(new ScoringFunctionFactory() {
         *
         * @Inject CadytsContext cadytsContext;
         *
         * @Inject ScoringParametersForPerson parameters;
         *
         * @Override public ScoringFunction createNewScoringFunction(Person person) {
         * final ScoringParameters params = parameters.getScoringParameters(person);
         *
         * SumScoringFunction scoringFunctionAccumulator = new SumScoringFunction();
         * scoringFunctionAccumulator.addScoringFunction(new
         * CharyparNagelLegScoring(params, controler.getScenario().getNetwork(),
         * config.transit().getTransitModes()));
         * scoringFunctionAccumulator.addScoringFunction(new
         * CharyparNagelActivityScoring(params)) ;
         * scoringFunctionAccumulator.addScoringFunction(new
         * CharyparNagelAgentStuckScoring(params));
         *
         * final CadytsScoring<Link> scoringFunction = new
         * CadytsScoring<>(person.getSelectedPlan(), config, cadytsContext);
         * scoringFunction.setWeightOfCadytsCorrection(30. *
         * config.planCalcScore().getBrainExpBeta()) ;
         * scoringFunctionAccumulator.addScoringFunction(scoringFunction );
         *
         * return scoringFunctionAccumulator; } }) ;
         */
        controler.addOverridingModule(new SwissRailRaptorModule());
        controler.run();
        // Temporary fix, will change later when we get emission files for the other cities
        EmissionsOffline.calculateEmissions(city, String.valueOf(simulationId), resultsPath);
        logger.info("EMISSIONS FINISHED\n\n\n");
        sendProgress(simulation, Simulation.Status.FINISHED, "Simulation run finished");

    }

    public Plan getPlanForSimulation(Simulation simulation) throws RuntimeException {
        Plan plan = simulation.getPlan();
        return plan;
    }

    public Path createResultDir(int simulationId) {
        Path resultsPath = Paths.get("data", this.city, "simulations", String.valueOf(simulationId), "results",
                DateTime.now().toString().replace(":", "-"));
        File resultsDir = new File(resultsPath.toString());
        if (!resultsDir.exists()) {
            if (!resultsDir.mkdirs()) {
                logger.error("Could not create result directory with path " + resultsPath);
            }
        }
        return resultsPath;
    }

    public void sendProgress(Simulation simulation, Simulation.Status status, String message) {
        simulation.setStatus(status);
        simulationRepository.save(simulation);
        this.simpMessagingTemplate.convertAndSend("/topic/simulationRunStatus", simulation,
                Collections.singletonMap("statusWSMessage", message));
    }


}
