package si.ijs.urbanite.simserver.travel_demand;

import org.matsim.api.core.v01.Coord;
import org.matsim.api.core.v01.Id;
import org.matsim.api.core.v01.Scenario;
import org.matsim.api.core.v01.network.Network;
import org.matsim.api.core.v01.population.*;
import org.matsim.core.api.internal.MatsimWriter;
import org.matsim.core.config.Config;
import org.matsim.core.config.ConfigUtils;
import org.matsim.core.scenario.ScenarioUtils;
import org.matsim.core.utils.geometry.CoordinateTransformation;
import org.matsim.core.utils.geometry.transformations.TransformationFactory;
import si.ijs.urbanite.simserver.simulation.SimulationWithParamsDTO;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Random;
import java.util.concurrent.atomic.AtomicLong;

public class TravelDemandGenerationHelsinkiCut {

    private static AtomicLong idCounter = new AtomicLong();

    public static String createID() {
        return String.valueOf(idCounter.getAndIncrement());
    }

    // create ferry arrivals departures
    public static void addFerries(String path, Population population, PopulationFactory populationFactory,
            Integer scaleFactor, String epsg) throws IOException {

        // scaleFactor for 10% population is 10

        // define locations of port, center and motorway
        ArrayList<Double> x = new ArrayList<>();
        ArrayList<Double> y = new ArrayList<>();

        // define coordinate transformation
        CoordinateTransformation transformation = TransformationFactory
                .getCoordinateTransformation(TransformationFactory.WGS84, epsg);

        double x_harbour = 60.15117250687363;
        double y_harbour = 24.915704166846062;
        double x_motorway = 60.16172281196161;
        double y_motorway = 24.8824963246887;
        double x_center = 60.16919848665633;
        double y_center = 24.933156410861443;

        String car = "car";
        String walk = "walk";
        String bicycle = "bicycle";

        ArrayList<String> modes = new ArrayList<>();
        modes.add(car);
        modes.add(walk);
        modes.add(bicycle);


        x.add(x_harbour);
        x.add(x_center);
        x.add(x_motorway);

        y.add(y_harbour);
        y.add(y_center);
        y.add(y_motorway);

        String category;
        int ferryCapacity;
        double time;

        Random random = new Random();
        // Read ferries data
        boolean header = true;
        BufferedReader bf = new BufferedReader(
                new FileReader(Paths.get("data", "helsinki", "before_simulation", "ferriesJatkasaari.csv").toString()));
        String line;
        while ((line = bf.readLine()) != null) {
            if (header) {
                header = false;
            } else {
                String[] row = line.split(",");
                category = row[6];
                time = Double.parseDouble(row[2]);
                ferryCapacity = Integer.parseInt(row[3]) / scaleFactor;

                // create pass count between 30% and 80% of the whole capacity
                int passCount = ferryCapacity * (random.nextInt(50) + 30) / 100; // bound of 50 integers, starting from
                                                                                 // 30

                int m = random.nextInt(3);
                if (row[6].equals("ARRIVAL")) {
                    for (int i = 0; i < passCount; i++) {
                        Person person = populationFactory.createPerson(Id.createPersonId("ferryPass" + createID()));
                        population.addPerson(person);
                        Plan plan = populationFactory.createPlan();

                        // first activity starts at the port in Jatkasaari
                        Activity activity1 = populationFactory.createActivityFromCoord("ferry",
                                transformation.transform(new Coord(y.get(0), x.get(0))));
                        activity1.setEndTime(time);
                        plan.addActivity(activity1);
                        plan.addLeg(populationFactory.createLeg(modes.get(m)));

                        // second activity is going on the motorway or to the center - chosen randomly
                        Activity activity2 = populationFactory.createActivityFromCoord("ferry", transformation
                                .transform(new Coord(y.get(random.nextInt(2) + 1), x.get(random.nextInt(2) + 1))));
                        activity2.setStartTime(time + random.nextInt(20 * 60) + 40 * 60);
                        activity2.setEndTime(24 * 60 * 60); // never come back
                        plan.addActivity(activity2);
                        plan.addLeg(populationFactory.createLeg(modes.get(m)));

                        // TODO add activity3 same as activity1 since all agents should start and end
                        // with same activity
                        Activity activity3 = populationFactory.createActivityFromCoord("ferry",
                                transformation.transform(new Coord(y.get(0), x.get(0))));
                        activity3.setEndTime(24 * 60 * 60);
                        plan.addActivity(activity3);

                        person.addPlan(plan);
                    }
                } else if (row[6].equals("DEPARTURE")) {
                    for (int i = 0; i < passCount; i++) {
                        Person person = populationFactory.createPerson(Id.createPersonId("ferryPass_" + createID()));
                        population.addPerson(person);
                        Plan plan = populationFactory.createPlan();

                        // first activity starts from the center or the motorway
                        Activity activity1 = populationFactory.createActivityFromCoord("ferry", transformation
                                .transform(new Coord(y.get(random.nextInt(2) + 1), x.get(random.nextInt(2) + 1))));
                        activity1.setEndTime(time - random.nextInt(60 * 60) + 60 * 60);
                        plan.addActivity(activity1);
                        plan.addLeg(populationFactory.createLeg(modes.get(m)));

                        // second activity is going to the port
                        Activity activity2 = populationFactory.createActivityFromCoord("ferry",
                                transformation.transform(new Coord(y.get(0), x.get(0))));
                        activity2.setStartTime(time);
                        activity2.setEndTime(24 * 60 * 60); // never come back
                        plan.addActivity(activity2);
                        plan.addLeg(populationFactory.createLeg(modes.get(m)));

                        // TODO add activity3 same as activity1 since all agents should start and end
                        // with same activity
                        person.addPlan(plan);
                        Activity activity3 = populationFactory.createActivityFromCoord("ferry", transformation
                                .transform(new Coord(y.get(random.nextInt(2) + 1), x.get(random.nextInt(2) + 1))));
                        activity3.setEndTime(24 * 60 * 60);
                        plan.addActivity(activity3);

                        person.addPlan(plan);
                    }
                }
            }
        }
    }

    public static void createActivities(Population population, PopulationFactory populationFactory, Integer scaleFactor,
            String epsg, ArrayList<Double> x, ArrayList<Double> y) throws IOException {
        // --------------------------------------
        // Age groups:
        // 1. 0-18
        // 2. 18-25
        // 3. 25-35
        // 4. 35-45
        // 5. 45-55
        // 6. 55-65
        // 7. 65+
        // -----------------------------------------

        Random random = new Random();
        CoordinateTransformation transformation = TransformationFactory
                .getCoordinateTransformation(TransformationFactory.WGS84, epsg);

        // define modes
        String car = "car";
        String walk = "walk";
        String pt = "pt";
        String bicycle = "bicycle";

        // define activity type
        String work = "work";
        String leisure = "leisure";
        String shopping = "shopping";
        String education = "education";
        String other = "other";

        // car availability
        String always = "always";
        String never = "never";
        String sometime = "sometime";

        // sex
        String male = "male";
        String female = "female";

        ArrayList<String> mode = new ArrayList<>(); // this array's types must be used as the factories from the
                                                    // framework require very specific input type
        mode.add(car);
        mode.add(bicycle);
        mode.add(walk);
        mode.add(pt);

        ArrayList<String> availability = new ArrayList<>();
        availability.add(always);
        availability.add(sometime);
        availability.add(never);

        ArrayList<Object> sex = new ArrayList<>();
        sex.add(male);
        sex.add(female);

        ArrayList<String> activity_first = new ArrayList<>();
        activity_first.add(work);
        activity_first.add(education);
        activity_first.add(leisure); // for retired people

        ArrayList<String> activity_second = new ArrayList<>();
        activity_second.add(leisure);
        activity_second.add(shopping);
        activity_second.add(other);

        // scale down the population to 10%
        ArrayList<Integer> index = new ArrayList<>();
        for (int i = 0; i < 639190 * scaleFactor / 100; i++) {
            index.add(random.nextInt(639190));
        }

        // read micrdata
        BufferedReader microdata = new BufferedReader(
                new FileReader(Paths.get("data", "helsinki", "before_simulation", "indHelsinki.csv").toString()));
        String line;
        boolean header = true; // check if we are in the header
        Integer counter = 0; // only select lines chosen by index array
        while ((line = microdata.readLine()) != null) {
            String[] person_attr = line.split(",");
            if (header) {
                header = false;
            } else {
                if (index.contains(counter)) {
                    String ageString = person_attr[1];
                    int age = Integer.parseInt(ageString);
                    String s = person_attr[2];
                    int home = Integer.parseInt(person_attr[3]);

                    // --------------------------------------------------
                    // define car availability
                    // --------------------------------------------------
                    int aw = 0; // car availability
                    if (age == 1) {
                        aw = 2;
                    } else {
                        aw = random.nextInt(2 + 1);
                    }

                    // -------------------------------------------------------
                    // define student, employee, retired
                    // -------------------------------------------------------
                    int rdE;
                    if (age == 1) {
                        rdE = 0; // student
                    } else if (age == 1 || age == 2) {
                        rdE = random.nextInt(2); // worker or student
                    } else if (age > 3 && age < 6) {
                        rdE = 1; // definitely worker
                    } else {
                        rdE = 2; // elderly
                    }

                    int m;
                    if (availability.get(aw).equals("never")) {
                        m = random.nextInt(2) + 1;
                    } else {
                        m = random.nextInt(3);
                    }

                    double blur_loc = -0.01 + (0.01 - (-0.01)) * random.nextDouble(); // in meters

                    Person person = populationFactory.createPerson(Id.createPersonId(person_attr[0]));
                    person.getAttributes().putAttribute("age group", age);
                    person.getAttributes().putAttribute("student, employee, retired", rdE);
                    person.getAttributes().putAttribute("sex", s);
                    person.getAttributes().putAttribute("car availability", availability.get(aw));

                    population.addPerson(person);

                    Plan plan = populationFactory.createPlan();

                    // -------------------------------------------------------------------------------------------------------------------
                    // Create the first activity. The first activity start time is different for
                    // students (young, old) | workers | retired
                    // --------------------------------------------------------------------------------------------------------------------
                    int start_time;
                    int end_time;

                    if (rdE == 0) {
                        // student
                        if (age == 1) {
                            start_time = 7 * 60 * 60 + random.nextInt(7200);
                            Activity activity1 = populationFactory.createActivityFromCoord("home", transformation
                                    .transform(new Coord(y.get(home - 1) + blur_loc, x.get(home - 1) + blur_loc)));
                            activity1.setEndTime(start_time - (random.nextInt(30 * 60) + 30 * 60));
                            plan.addActivity(activity1);
                            plan.addLeg(populationFactory.createLeg(mode.get(m)));
                            Activity activity2 = populationFactory.createActivityFromCoord("education",
                                    transformation.transform(new Coord(y.get(random.nextInt(8)) + blur_loc,
                                            x.get(random.nextInt(8)) + blur_loc)));
                            activity2.setStartTime(start_time); // start time from 07.00 to 09.00
                            activity2.setEndTime(start_time + random.nextInt(4 * 60 * 60) + 4 * 60 * 60);
                            plan.addActivity(activity2);
                            plan.addLeg(populationFactory.createLeg(mode.get(m)));
                        } else if (age == 2) {
                            // if the person is old student
                            start_time = 7 * 60 * 60 + random.nextInt(7200); // start time of first activity
                            Activity activity1 = populationFactory.createActivityFromCoord("home", transformation
                                    .transform(new Coord(y.get(home - 1) + blur_loc, x.get(home - 1) + blur_loc)));
                            activity1.setEndTime(start_time - (random.nextInt(30 * 60) + 30 * 60));
                            plan.addActivity(activity1);
                            plan.addLeg(populationFactory.createLeg(mode.get(m)));
                            Activity activity2 = populationFactory.createActivityFromCoord("education",
                                    transformation.transform(new Coord(y.get(random.nextInt(8)) + blur_loc,
                                            x.get(random.nextInt(8)) + blur_loc)));
                            end_time = start_time + random.nextInt(4 * 60 * 60) + 6 * 60 * 60;
                            activity2.setStartTime(start_time); // start time from 07.00 to 09.00
                            activity2.setEndTime(end_time);
                            plan.addActivity(activity2);
                            plan.addLeg(populationFactory.createLeg(mode.get(m)));

                            // old students have leisure activities before midnight
                            Activity act_other = populationFactory.createActivityFromCoord(
                                    activity_second.get(random.nextInt(3)),
                                    transformation.transform(new Coord(y.get(random.nextInt(8)) + blur_loc,
                                            x.get(random.nextInt(8)) + blur_loc)));
                            start_time = end_time + 60 * 60;
                            end_time = start_time + random.nextInt(4 * 60 * 60);
                            act_other.setStartTime(start_time);
                            act_other.setEndTime(end_time);
                            plan.addActivity(act_other);
                            plan.addLeg(populationFactory.createLeg(mode.get(m)));
                        }
                    }
                    // -------------------------------------------------------------------------------------------------------------------
                    // Create the first activity for workers
                    // --------------------------------------------------------------------------------------------------------------------
                    else if (rdE == 1) {
                        // regular workers, first activity is work, then leisure/shopping
                        start_time = 6 * 60 * 60 + random.nextInt(3 * 60 * 60);
                        Activity activity1 = populationFactory.createActivityFromCoord("home", transformation
                                .transform(new Coord(y.get(home - 1) + blur_loc, x.get(home - 1) + blur_loc)));
                        activity1.setEndTime(start_time - (random.nextInt(40 * 60) + 40));
                        plan.addActivity(activity1);
                        plan.addLeg(populationFactory.createLeg(mode.get(m)));
                        end_time = start_time + random.nextInt(3 * 60 * 60) + 6 * 60 * 60;
                        Activity activity2 = populationFactory.createActivityFromCoord("work", transformation.transform(
                                new Coord(y.get(random.nextInt(8)) + blur_loc, x.get(random.nextInt(8)) + blur_loc)));
                        activity2.setStartTime(start_time);
                        activity2.setEndTime(end_time);
                        plan.addActivity(activity2);
                        plan.addLeg(populationFactory.createLeg(mode.get(m)));

                        // after work can have two other activities (shopping, leisure, other)
                        for (int i = 0; i < random.nextInt(2) + 1; i++) {
                            start_time = end_time + random.nextInt(1 * 60 * 60) + 30 * 60;
                            end_time = start_time + random.nextInt(1 * 60 * 60);
                            Activity act_other = populationFactory.createActivityFromCoord(
                                    activity_second.get(random.nextInt(3)),
                                    transformation.transform(new Coord(y.get(random.nextInt(8)) + blur_loc,
                                            x.get(random.nextInt(8)) + blur_loc)));
                            act_other.setStartTime(start_time);
                            act_other.setEndTime(end_time);
                            plan.addActivity(act_other);
                            plan.addLeg(populationFactory.createLeg(mode.get(m)));
                        }

                    }
                    // -------------------------------------------------------------------------------------------------------------------
                    // Create the first activity for retired
                    // --------------------------------------------------------------------------------------------------------------------
                    else {
                        // retired, can have up to 3 activities
                        start_time = 9 * 60 * 60 + random.nextInt(2 * 60 * 60);
                        Activity activity1 = populationFactory.createActivityFromCoord("home", transformation
                                .transform(new Coord(y.get(home - 1) + blur_loc, x.get(home - 1) + blur_loc)));
                        activity1.setEndTime(start_time - random.nextInt(40 * 60));
                        plan.addActivity(activity1);
                        plan.addLeg(populationFactory.createLeg(mode.get(m)));
                        for (int i = 0; i < random.nextInt(3) + 1; i++) {
                            end_time = start_time + random.nextInt(40 * 60) + 40 * 60;
                            Activity act_other = populationFactory.createActivityFromCoord(
                                    activity_second.get(random.nextInt(3)),
                                    transformation.transform(new Coord(y.get(random.nextInt(8)) + blur_loc,
                                            x.get(random.nextInt(8)) + blur_loc)));
                            act_other.setStartTime(start_time);
                            act_other.setEndTime(end_time);
                            plan.addActivity(act_other);
                            plan.addLeg(populationFactory.createLeg(mode.get(m)));
                            start_time = end_time + random.nextInt(2 * 60 * 60) + 30 * 60;
                        }
                    }
                    Activity activity3 = populationFactory.createActivityFromCoord("home", transformation
                            .transform(new Coord(y.get(home - 1) + blur_loc, x.get(home - 1) + blur_loc)));
                    plan.addActivity(activity3);
                    person.addPlan(plan);
                }
                counter++;
            }
        }

    }

    public static void generateTravelDemand(String net_id, String plan_id) throws IOException {

        // -------------------------------------------------------------------------------------------------------------------
        // Create list of x and y coordinates for 8 districts:
        // 1 Eteläinen","2 Läntinen","3 Keskinen","4 Pohjoinen","5 Koillinen","6
        // Kaakkoinen","7 Itäinen","8 Östersundom
        // --------------------------------------------------------------------------------------------------------------------

        //center
        double x1 = 60.16754792966954;
        double y1 = 24.931060272122014;
//        double x2 = 60.189765137160265;
//        double y2 = 24.949102572621623;
//        double x3 = 60.186179036620985;
//        double y3 = 24.94955358473212;
//        double x4 = 60.24781474378347;
//        double y4 = 24.948682371492083;
//        double x5 = 60.2509387705292;
//        double y5 = 25.012269127901845;
//        double x6 = 60.19427260811648;
//        double y6 = 25.034194165509337;
        double x7 = 60.15806699873942;
        double y7 = 24.95615737066421;
//        double x8 = 60.25155543103413;
//        double y8 = 25.17263932194783;

        ArrayList<Double> x = new ArrayList<>();
        x.add(x1);
//        x.add(x2);
//        x.add(x3);
//        x.add(x4);
//        x.add(x5);
//        x.add(x6);
        x.add(x7);
//        x.add(x8);

        ArrayList<Double> y = new ArrayList<>();
        y.add(y1);
//        y.add(y2);
//        y.add(y3);
//        y.add(y4);
//        y.add(y5);
//        y.add(y6);
        y.add(y7);
//        y.add(y8);

        Config config = ConfigUtils
                .loadConfig(Paths.get("data", "helsinki", "original_input", "config.xml").toString());
        Scenario sc = ScenarioUtils.createScenario(config);

        Population population = sc.getPopulation();
        PopulationFactory populationFactory = population.getFactory();

        Network network = sc.getNetwork();

        String ferriesPath = Paths.get("data", "helsinki", "before_simulation", "ferriesJatkasaari.csv").toString();

        addFerries(ferriesPath, population, populationFactory, 10, "epsg:32634");
//        createActivities(population, populationFactory, 1, "epsg:32634", x, y);

        MatsimWriter popWriter = new PopulationWriter(population, network);
        popWriter.write(
                Paths.get("data", "helsinki", "original_input", "networks", net_id, "plans", plan_id, "plans.xml")
                        .toString());
    }

    public static void main(String[] args) throws IOException {
//        generateTravelDemand("1","1");
    }

}