package si.ijs.urbanite.simserver.utils;

import org.jfree.data.json.impl.JSONArray;
import org.jfree.data.json.impl.JSONObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import tech.tablesaw.api.Table;
import tech.tablesaw.columns.Column;
import tech.tablesaw.io.DataFrameReader;
import tech.tablesaw.io.ReaderRegistry;
import tech.tablesaw.io.json.JsonReader;
import tech.tablesaw.io.Source;

import javax.annotation.Nullable;
import java.io.*;
import java.util.*;

public class GeoJsonExport {

    private static final Logger logger = LoggerFactory.getLogger("GeoJsonExport");

    public static JSONObject networkToGeoJson(File networkFile) {
        SaxParser.parseXML(networkFile);
        Map<String, List<List<Double>>> streets = StreetsHandler.streets;

        JSONObject geojson = new JSONObject();
        geojson.put("type", "FeatureCollection");

        JSONArray features = new JSONArray();
        for (String streetName : streets.keySet()) {
            JSONObject geometry = createGeometryLineString(streets.get(streetName));
            JSONObject properties = createProperties(streetName);
            JSONObject feature = createFeature(properties, geometry);
            features.add(feature);
        }
        geojson.put("features", features);

        logger.debug("Network GeoJson created.");
        return geojson;
    }

    private static JSONObject createFeature(JSONObject properties, JSONObject geometry) {
        JSONObject feature = new JSONObject();
        feature.put("type", "Feature");
        feature.put("properties", properties);
        feature.put("geometry", geometry);

        logger.debug("GeoJson Feature created.");
        return feature;
    }

    private static JSONObject createProperties(String streetName) {
        JSONObject properties = new JSONObject();
        properties.putAll(new HashMap<String, String>() {{
            put("streetName", streetName);
            // TODO: add to StreetHandler capacity map, speed map etc.
        }});

        logger.debug("GeoJson Properties created.");
        return properties;
    }

    private static JSONObject createGeometryLineString(List<List<Double>> streetCoordinates) {
        JSONObject geometry = new JSONObject();
        geometry.put("type", "LineString");
        JSONArray coordinates = new JSONArray();
        coordinates.addAll(streetCoordinates);
        geometry.put("coordinates", coordinates);

        logger.debug("GeoJson Geometry created.");
        return geometry;
    }

    public static JSONObject createDistrictsGeoJson(List<File> files) {
        // read data into tables
        List<Table> tables = readFiles(files);
        Table merged = mergeDistrictData(tables);
        return null;
    }

    private static List<Table> readFiles(List<File> files) {
        List<Table> tables = new ArrayList<>(files.size());
        //for (File f : files) {
        for (ListIterator<File> iter = files.listIterator(); iter.hasNext(); ) {
            File f = iter.next();

            if (f.getName().endsWith(".csv")) try {
                // csv files are slightly more annoying due to 2 row header etc
                //
                tables.add(Table.read().csv(f.getAbsolutePath()));
            } catch (IOException e) {
                handleIOException(f);
            } catch (Exception e) {
                // special case: remove first row/line and re-read
                if (f.getName().endsWith("numero_habitantes_distrito_barrio_estudios_2020.csv")) {
                    try {
                        File fixedFile = removeFirstLine(f);
                        iter.add(fixedFile);
                    } catch (IOException ioe) {
                        System.out.println("could not fix file " + f.getAbsolutePath());
                    }
                }
                // failure :(
                System.out.println("failed reading file " + f.getAbsolutePath());
            }
            else System.out.println("ignoring file: " + f.getAbsolutePath());
        }
        return tables;
    }

    @Nullable
    private static File removeFirstLine(File f) throws IOException {
        File tempFile = new File(f.getAbsolutePath() + ".tmp");
        BufferedReader br = null;
        try {
            br = new BufferedReader(new FileReader(f));
        } catch (FileNotFoundException e) {
            System.out.println("Trying to fix the file, file not found: " + f.getAbsolutePath());
            return null;
        }
        PrintWriter pw = null;
        try {
            pw = new PrintWriter(new FileWriter(tempFile));
        } catch (IOException e) {
            System.out.println("Trying to fix the file, could not create temp file.");
            return null;
        }

        String line;
        int count = 0;
        while ((line = br.readLine()) != null) {
            if (count == 0) continue;
            count++;
            pw.println(line);
        }
        pw.flush();
        pw.close();
        br.close();
        if (!f.delete())
            System.out.println("Could not delete original file! Big error");
        if (!tempFile.renameTo(f))
            System.out.println("Could not rename temp file! Big error");
        return new File(tempFile.getAbsolutePath());
    }

    private static Table mergeDistrictData(List<Table> tables) {
        Table districts = tables.stream()
                .filter(table -> table.name().equals("barios.csv"))
                .findFirst()
                .get(); // will be present! no need to isPresent()
        tables.remove(districts);
        String idColName = "TBAR_COD_BARRIO";
        String idColNumInhabitantsSex = "KODEA/CODIGO";
        String idColNumInhabitantsAge = "COD. BARRIO"; // first line additional info
        String idColNumInhabitantsEducation = "BARRIO";
        // thank you for this perfect matching of id col names
        // TODO figure out this in python or excel!
        for (Table t : tables) {
        }
        return null;
    }

    /**
     * Prints the absolute of a file that could not be opened.
     *
     * @param f File object that threw IOException on read.
     */
    private static void handleIOException(File f) {
        System.out.println("File not found: " + f.getAbsolutePath());
    }

}
