package si.ijs.urbanite.simserver.utils;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.xml.sax.Attributes;
import org.xml.sax.helpers.DefaultHandler;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class StreetsHandler extends DefaultHandler {

    private final Logger logger = LoggerFactory.getLogger(this.getClass());

    public static Map<String, List<List<Double>>> streets;
    public static Map<Long, List<Double>> nodeCoordinates;
    public static Map<Long, List<Long>> linkNodes;

    private static Map<String, List<Long>> streetsNodes;
    private static Long lastLink;
    private static StringBuilder attributeValue;
    private static String attributeName = "-";

    @Override
    public void startDocument() {
        streetsNodes = new HashMap<>();
        nodeCoordinates = new HashMap<>();
        linkNodes = new HashMap<>();
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes attributes) {
        switch (qName.toLowerCase()) {
            case "node":
                handleNode(attributes);
                break;
            case "link":
                handleLink(attributes);
                break;
            case "attribute":
                handleAttribute(attributes);
                break;
            default:
                break;
        }
    }

    @Override
    public void endElement(String uri, String localName, String qName) {
        switch (qName.toLowerCase()) {
            case "nodes":
            case "node":
            case "link":
                break;
            case "attribute":
                endAttribute();
                break;
            case "links":
                normalizeStreets();
                normalizeStreetsToCoordinates();
                break;
        }
    }

    @Override
    public void characters(char[] ch, int start, int length) {
        attributeValue.append(ch, start, length);
    }

    /**
     * Convert each entry of streets from array of nodes to array of coordinate pairs
     */
    private void normalizeStreetsToCoordinates() {
        streets = new HashMap<>();
        for (String streetName : streetsNodes.keySet()) {
            List<Long> streetNodes = streetsNodes.get(streetName);
            List<List<Double>> streetCoordinates = new ArrayList<>();
            for (long node : streetNodes) {
                streetCoordinates.add(nodeCoordinates.get(node));
            }
            streets.put(streetName, streetCoordinates);
        }
    }

    /**
     * Convert each entry of streets from array of links to array of nodes
     */
    private void normalizeStreets() {
        Map<String, List<Long>> normalizedStreets = new HashMap<>();
        for (String streetName : streetsNodes.keySet()) {
            // streets are list of links
            List<Long> streetLinks = streetsNodes.get(streetName);
            // streets will be list of nodes
            List<Long> streetNodes = new ArrayList<>(streetLinks.size() + 1);
            // store previous for compare
            long previousLink = streetLinks.get(0);
            for (long link : streetLinks) {
                // skip first link - first node will be added after loop for simplicity
                if (previousLink == link) continue;
                // get both links' nodes
                List<Long> previousNodes = linkNodes.get(previousLink);
                List<Long> thisNodes = linkNodes.get(link);
                // choose node to add to street by finding equal nodes in both neighbouring links
                long nodeToAdd;
                if (previousNodes.get(0).equals(thisNodes.get(0))) {
                    nodeToAdd = thisNodes.get(0);
                } else if (previousNodes.get(0).equals(thisNodes.get(1))) {
                    nodeToAdd = thisNodes.get(1);
                } else {
                    logger.error("Nodes in street not connected: " +
                            "previous link: " + previousLink +
                            "current link: " + link);
                    throw new RuntimeException("Nodes in street not connected.");
                }
                streetNodes.add(nodeToAdd);
            } 
            // add first node - node from first link that is currently not the first in streetNodes
            List<Long> firstLink = linkNodes.get(streetLinks.get(0));
            streetNodes.add(
                    0,
                    streetNodes.get(0).equals(firstLink.get(0)) ? firstLink.get(1) : firstLink.get(0));
            normalizedStreets.put(streetName, streetNodes);
        }
        // replace streets
        streetsNodes = normalizedStreets;
    }

    /**
     * Start Node element: store node id with coordinates
     *
     * @param attributes SAX parser attributes
     */
    private void handleNode(Attributes attributes) {
        // ignore pt virtual links
        long nodeId;
        try {
            nodeId = Long.parseLong(attributes.getValue("id"));
        } catch (NumberFormatException e) {
            return;
        }
        // get coords and store to node
        double x = Double.parseDouble(attributes.getValue("x"));
        double y = Double.parseDouble(attributes.getValue("y"));
        List<Double> coords = new ArrayList<>(Arrays.asList(x, y));
        nodeCoordinates.put(nodeId, coords);
    }

    /**
     * Start Attribute element: if street name, store links under street name
     *
     * @param attributes SAX parser attributes
     */
    private void handleAttribute(Attributes attributes) {
        attributeValue = new StringBuilder();
        for (int i = 0; i < attributes.getLength(); i++) {
            String attrName = attributes.getValue(i);
            if (attrName.equals("osm:way:name")) attributeName = attrName;
        }
    }

    /**
     * End attribute and stores value
     */
    private void endAttribute() {
        if (attributeName.equals("osm:way:name")) {
            String attrValue = attributeValue.toString();
            List<Long> streetLinks;
            if (streetsNodes.containsKey(attrValue)) streetLinks = streetsNodes.get(attrValue);
            else streetLinks = new ArrayList<>();
            streetLinks.add(lastLink);
            streetsNodes.put(attrValue, streetLinks);
        }
        attributeName = "-";
    }

    /**
     * Start Link element: read link id and node ids and store
     *
     * @param attributes SAX parser attributes
     */
    private void handleLink(Attributes attributes) {
        long linkId = Long.parseLong(attributes.getValue("id"));
        long nodeId1 = Long.parseLong(attributes.getValue("from"));
        long nodeId2 = Long.parseLong(attributes.getValue("to"));
        List<Long> nodes = Arrays.asList(nodeId1, nodeId2);
        linkNodes.put(linkId, nodes);
        lastLink = linkId;
    }
}
